unit myDebug;

interface

uses
  Windows, SysUtils, cadt, myPe, myTypes, Messages;

CONST
  version = '0.4';

TYPE TMyHandler = procedure(const ERec : _EXCEPTION_RECORD);
TYPE TMyLogHandler = procedure(const sLog : string);

TYPE TDisasmOptions = packed record
  Len     : DWORD;
  Command : TDisCommand;
  Instr   : TInstruction;
  Mnem    : string[150];
END;

TYPE PDisasmOptions = ^TDisasmOptions;

TYPE TBPX = packed record
  addr : dword;
  bb   : byte;
END;

TYPE TBPH = packed record
  addr  : dword;
  dtype : dword;
END;

TYPE TBPM = packed record
  addr : dword;
  size : dword;
  oldp : dword;
END;

TYPE
  PDebugger = ^TDebugger;
  TDebugger = class (TObject)

private
  PEB        : PPEBNT;
  FDebugLog  : string;
  iNeedStop  : integer;
  iState     : integer;
  iTracing   : integer;
  DbgControl : dword;
  ImageBase  : dword;
  EntryPoint : dword;
  PID        : dword;

  FEventReason  : dword;
  FBpxIndex     : integer;

  CustomHandler    : TMyHandler;
  CustomLogHandler : TMyLogHandler;

  procedure HandleMyBpx();
  procedure SystemBreak(const ERec : _EXCEPTION_RECORD);
  procedure ExceptionSingleStep(const ERec : _EXCEPTION_RECORD);
  procedure ExceptionGuardPage (const ERec : _EXCEPTION_RECORD);
  procedure ExceptionBreakPoint(const ERec : _EXCEPTION_RECORD);
  procedure HideDBG();
  procedure CreateProcessDebugEvent(const DE : TDebugEvent);
  procedure ProcessLoadDllDebugEvent(const DE : TDebugEvent);

  function AllocMem():pointer;
  function PrepareDebug():integer;
  function ReadMemorySelf(dwAddress:DWORD;pBuffer:Pointer;dwSize:dword):integer;
  function GetErrorCode(ErrCode: dword): string;

  function MyBPX(const dwAdress:DWORD): integer;
  function MyBPH(const dwAdress:DWORD): integer;
  function MyBPM(const dwAdress:DWORD): integer;

  //function TraceToOep():DWORD;

  procedure Log(const sLog: string);
  procedure ProcessOutputDebugString(const DE : TDebugEvent);
  
published
//
public
// USER HANDLER'S ----------
  DllHandler        : dword;
  Int3Handler       : dword;
  PageGuardHandler  : dword;
  SingleStepHandler : dword;
  SystemBpHandler   : dword;
  OepHandler        : dword;
  LogHandler        : dword;
// -------------------------
  iSetEpBreak : integer;

  iHideDdg    : integer;
  iHandleDll  : integer;
  iLastError  : integer;
  iLogToFile  : integer;
  iOepTrace   : integer;
  sLogName    : string;
  dwOEP       : dword;

  lpThreadLocalBase : pointer;
// -------------------------
  si: tStartupInfo;
  pi: tProcessInformation;

  function WriteMemory(dwAddress:DWORD;pBuffer:Pointer;dwSize:dword):integer;
  function ReadMemory(const dwAddress: DWORD; const pBuffer: Pointer; const dwSize: dword):integer;
  procedure Disasm(const ADDR, RVA: dword; pDisOpt: PDisasmOptions);

  function GetModuleSize(dwModuleHandle: DWord): DWord;
  function SetHardwareBreak(const iNumber:integer;dwAddress:DWORD):Integer;
  function SetBpx(const iNumber:integer;dwAddress:DWORD):Integer;
  function DumpProcess():integer;
  function DelBpx(const iNumber:integer):Integer;
  function GetEIP():DWORD;
  function GetESP():DWORD;
  function GetFlags():DWORD;
  function GetRegister(const num:integer):DWORD;
  function GetDebugRegister(const num:integer):DWORD;
  function SetBPM(dwAddress,dwSize,dwNum:dword):boolean;
  function DelBPM(dwNum:dword):boolean;
  function DelHardwareBreak(const iNumber:integer):Integer;
  function DoDebugLoop(var dwState, dwReason: dword):dword;

  function GetPeb(lpThreadLocalBase:pointer): pointer;
  procedure SetTraceMode(const iTrace: integer);
  procedure ProcessTerminate();
  procedure ProcessSuspend();
  procedure ProcessResume();
  function RunDebug(const FileName:string): dword;
  procedure TerminateDebug();
 // procedure GetEax();

  property sDebugLog     : string read FDebugLog;
  //property dwEventReason : dword read FEventReason;
  property BpxIndex      : integer read FBpxIndex;
// -----------------------------------------------------------------------------
  constructor Create;
  destructor  Destroy; override;
// -----------------------------------------------------------------------------
END;

CONST
  MAX_COMMAND_COUNT = 1500;

  DEBUG_EVENT            = 1;

  CAN_NOT_CREATE_PROCESS = 500;
  CAN_NOT_GET_MEM        = 501;
  CAN_NOT_CREATE_FILE    = 502;
  CAN_NOT_WRITE_TO_FILE  = 503;
  CAN_NOT_OPEN_FILE      = 504;
  CAN_NOT_READ_MEM       = 505;
  NOT_VALID_PE_FILE      = 506;
  ERROR_SUCCES           = 507;
  BAD_HANDLER_INT3       = 508;
  BAD_HANDLER_GUARD      = 509;
  BAD_HANDLER_STEP       = 510;
  BAD_HANDLER_SYS        = 511;
  BAD_HANDLER_OEP        = 512;

  PROCESS_TERMINATED = 400;
  PROCESS_RUNNING    = 401;
  PROCESS_SUSPENDED  = 402;
  
  TRACE_FLAG = $100;
// DRX -------------------------------------------------------------------------
  BPM0_ENABLED    = $00000001;
  BPM0_W          = $00010000;
  BPM0_RW         = $00030000;
  BPM0_1b         = $00000000;
  BPM0_2b         = $00040000;
  BPM0_4b         = $000C0000;
  BPM0_nb         = $00080000;
  BPM1_ENABLED    = $00000004;
  BPM1_W          = $00100000;
  BPM1_RW         = $00300000;
  BPM2_ENABLED    = $00000010;
  BPM2_W          = $01000000;
  BPM2_RW         = $03000000;
  BPM3_ENABLED    = $00000040;
  BPM3_W          = $10000000;
  BPM3_RW         = $30000000;
  BPM_LOCAL_EXACT = $00000100;

  EP_BREAK        = 0;
  EP_HANDLER      = $10000000;
  BPX_HANDLER     = $20000000;
  BPX_SYS_HANDLER = $21000000;
  BPH_HANDLER     = $30000000;
  BPM_HANDLER     = $40000000;
  TRACING_HANDLER = $50000000;
// ----------------------------------------------------------------------------

implementation

VAR
  pe   : TPEEditor;
  Cont : TContext;
  DE   : TDebugEvent;
  FilePath : string;

  BPH      : array [0..3]  of TBPH;
  BPX      : array [0..15] of TBPX;
  BPM      : array [0..15] of TBPM;
  AddressX : array [0..50] of DWORD;

  bPatchPEB: boolean;
  pDbgBreakPoint:Pointer;
  IsFirstException:Boolean;

  tCommand : TDisasmOptions;
// ----------------------------------------------------------------------------
function GetModuleImageSize(const hmodule: dword):dword;
var
  PE_HEADEROFFSET: dword;
begin
  Move(Pointer(hmodule + $3C)^,PE_HEADEROFFSET,4);
  Move(Pointer(hmodule + PE_HEADEROFFSET + $50)^,Result,4);
end;

function GetModuleCodeSize(const hmodule: dword):dword;
var
  PE_HEADEROFFSET: dword;
begin
  Move(Pointer(hmodule + $3C)^,PE_HEADEROFFSET,4);
  Move(Pointer(hmodule + PE_HEADEROFFSET + $1C)^,Result,4);
end;

function TDebugger.GetPeb(lpThreadLocalBase:pointer): pointer;
var
  dwAddr : DWORD;
  PEBNT  : PPEBNT;
  dwTemp : dword;
begin
  dwAddr := DWORD(lpThreadLocalBase) + $30;
  dwTemp := 0;
  ReadMemory(dwAddr,@dwTemp,4);
  If (dwTemp > 0) then
  begin
    ReadMemory(dwTemp, PEBNT, SizeOf(TPEBNT));
    result := PEBNT;
  end;
end;

function TDebugger.GetErrorCode(ErrCode: dword): string;
begin
   case ErrCode of
      EXCEPTION_ACCESS_VIOLATION          : Result := 'Access Violation';
      EXCEPTION_ARRAY_BOUNDS_EXCEEDED     : Result := 'Array Bounds Exceeded';
      EXCEPTION_BREAKPOINT                : Result := 'Breakpoint';
      EXCEPTION_DATATYPE_MISALIGNMENT     : Result := 'Datatype Misalignment';
      EXCEPTION_FLT_DENORMAL_OPERAND      : Result := 'Flt Denormal Operand';
      EXCEPTION_FLT_DIVIDE_BY_ZERO        : Result := 'Float Divide by Zero';
      EXCEPTION_FLT_INEXACT_RESULT        : Result := 'Float Inexact Result';
      EXCEPTION_FLT_INVALID_OPERATION     : Result := 'Float Invalid Operation';
      EXCEPTION_FLT_OVERFLOW              : Result := 'Float Overflow';
      EXCEPTION_FLT_STACK_CHECK           : Result := 'Float Stack Check';
      EXCEPTION_FLT_UNDERFLOW             : Result := 'Float Underflow';
      EXCEPTION_ILLEGAL_INSTRUCTION       : Result := 'Illegal Instruction';
      EXCEPTION_IN_PAGE_ERROR             : Result := 'In Page Error';
      EXCEPTION_INT_DIVIDE_BY_ZERO        : Result := 'Int Divide by Zero';
      EXCEPTION_INT_OVERFLOW              : Result := 'Int Overflow';
      EXCEPTION_INVALID_DISPOSITION       : Result := 'Invalid Disposition';
      EXCEPTION_NONCONTINUABLE_EXCEPTION  : Result := 'Noncontinuable Exception';
      EXCEPTION_PRIV_INSTRUCTION          : Result := 'Privlidged Instruction';
      EXCEPTION_SINGLE_STEP               : Result := 'Single Step';
      EXCEPTION_STACK_OVERFLOW            : Result := 'Stack Overflow';
      EXCEPTION_GUARD_PAGE                : Result := 'Guard Page';
      EXCEPTION_INVALID_HANDLE            : Result := 'Invalid Handle';
      //DBG_CONTROL_C                       : Result := 'Control-C';
      else Result := 'Unknown Error';
   end;
end;

constructor TDebugger.Create();
begin
  inherited;

  Int3Handler       := 0;
  PageGuardHandler  := 0;
  SingleStepHandler := 0;

  iSetEpBreak  := 0;
  OepHandler   := 0;
  iOepTrace    := 0;
  iHandleDll   := 0;

  PE := TPEEditor.Create(nil);
  // ---
end;

Destructor TDebugger.Destroy;
begin
  //PE.Destroy;
  // ---
  inherited;
end;

Function TDebugger.GetDebugRegister(const num:integer):DWORD;
begin
  result:=$0BADC0DE;
  Cont.ContextFlags:=CONTEXT_DEBUG_REGISTERS;
  GetThreadContext(pi.hThread,cont);
  case num of
  	0: result:=cont.Dr0;
  	1: result:=cont.Dr1;
  	2: result:=cont.Dr2;
  	3: result:=cont.Dr3;
    6: result:=cont.Dr6;
    7: result:=cont.Dr7;
  end;
end;

Function TDebugger.GetRegister(const num:integer):DWORD;
begin
  result:=$0BADC0DE;
  Cont.ContextFlags:=CONTEXT_INTEGER;
  GetThreadContext(pi.hThread,cont);
  case num of
  	1: result:=cont.Eax;
  	2: result:=cont.Ecx;
  	3: result:=cont.Edx;
  	4: result:=cont.Ebx;
  	5: result:=cont.Esi;
  	6: result:=cont.Edi;
  end;
end;

function TDebugger.GetFlags():DWORD;
begin
  Cont.ContextFlags:=CONTEXT_CONTROL;
  GetThreadContext(pi.hThread,cont);
  result:=cont.EFlags;
end;

function TDebugger.GetESP():DWORD;
begin
  Cont.ContextFlags:=CONTEXT_CONTROL;
  GetThreadContext(pi.hThread,cont);
  result:=cont.Esp;
end;

function TDebugger.GetEIP():DWORD;
begin
  cont.Eip:=0;
  Cont.ContextFlags:=CONTEXT_CONTROL;
  GetThreadContext(pi.hThread,cont);
  result:=cont.Eip;
end;

procedure TDebugger.SetTraceMode(const iTrace: integer);
begin
  iTracing          := iTrace;
  Cont.ContextFlags := CONTEXT_CONTROL;
  GetThreadContext(pi.hThread,cont);
  If (iTrace = 0) then cont.EFlags := cont.EFlags xor TRACE_FLAG;
  If (iTrace = 1) then cont.EFlags := cont.EFlags or  TRACE_FLAG;
  SetThreadContext(pi.hThread, cont);
end;

function TDebugger.DelBPM(dwNum:dword):boolean;
var
  dwOld, dwFlag:dword;
begin
  result := False;
  If ((dwNum >= 0) and (dwNum <= 15) and (BPM[dwNum].addr > 0)) then
  begin
    result := VirtualProtectEx(pi.hProcess, POINTER(BPM[dwNum].addr), BPM[dwNum].size, dwFlag, BPM[dwNum].oldp);
    BPM[dwNum].addr := 0;
    BPM[dwNum].size := 0;
    BPM[dwNum].oldp := 0;
  end;
end;

function TDebugger.SetBPM(dwAddress,dwSize,dwNum:dword):boolean;
var
  dwOld,dwFlag:dword;
begin
  result:=false;
  If (dwNum < 0) or (dwNum > 15) then exit;
  BPM[dwNum].addr:=dwAddress;
  BPM[dwNum].size:=dwSize;
  dwFlag:=PAGE_EXECUTE_READWRITE or PAGE_GUARD;
  result:=VirtualProtectEx(pi.hProcess,POINTER(dwAddress),dwSize,dwFlag,dwOld);
  BPM[dwNum].size:=dwSize;
  BPM[dwNum].oldp:=dwOld;
end;

function TDebugger.AllocMem():pointer;
begin
  result:=VirtualAllocEx(pi.hProcess,nil,$1000,MEM_COMMIT or MEM_RESERVE,PAGE_EXECUTE_READWRITE);
end;

procedure TDebugger.HideDBG();
var
  dwAddr,dwTemp,zero:dword;
  pMem:pointer;
  fs30:dword;
begin
  zero   := 0;
  dwAddr := DWORD(lpThreadLocalBase) + $30;
  ReadProcessMemory(pi.hProcess,POINTER(dwAddr),@fs30,4,dwTemp);
  dwAddr := fs30 + $2;
  WriteProcessMemory(pi.hProcess,POINTER(dwAddr),@zero,1,dwTemp);
  dwAddr := fs30 + $68;
  WriteProcessMemory(pi.hProcess,POINTER(dwAddr),@zero,4,dwTemp);
  bPatchPEB := TRUE;
end;

procedure TDebugger.Log(const sLog: string);
begin
  If (LogHandler > 0) then
  begin
    @CustomLogHandler := POINTER(LogHandler);
    FDebugLog := sLog;
    CustomLogHandler(FDebugLog);
  end;
end;

function TDebugger.PrepareDebug():integer;
begin
  result    := 0;
  FDebugLog := 'Runing: ' + FilePath;
  Log(FDebugLog);
  ZeroMemory(@si,SizeOf(TStartupInfo));
  ZeroMemory(@pi,SizeOf(TProcessInformation));
  si.cb := SizeOf(TStartupInfo);
  //GetStartupInfo(si);
  //debugbreak;
  if (not CreateProcess(nil,PChar(FilePath),nil,nil,FALSE,CREATE_SUSPENDED or DEBUG_ONLY_THIS_PROCESS,nil,nil,si,pi)) then
  begin
    FDebugLog := 'Failed to create process';
    Log(FDebugLog);
    Exit;
  end;
  {if (not DebugActiveProcess(pi.dwProcessId)) then
  begin
    FDebugLog := '[Failed DebugActiveProcess]';
    Exit;
  end;}
  pDbgBreakPoint := GetProcAddress(GetModuleHandle(ntdll),'DbgBreakPoint');
  If (pDbgBreakPoint = nil) then
  begin
    pDbgBreakPoint := GetProcAddress(GetModuleHandle(kernel),'DebugBreak');
  end;
  IsFirstException := TRUE;
  iState := PROCESS_SUSPENDED;
  result := 1;
end;

procedure TDebugger.ProcessSuspend();
begin
  iState := PROCESS_SUSPENDED;
  SuspendThread(pi.hThread);
end;

procedure TDebugger.ProcessResume();
begin
  iState := PROCESS_RUNNING;
  ResumeThread(pi.hThread);
end;

procedure TDebugger.ProcessTerminate();
begin
  iState := PROCESS_TERMINATED;
  TerminateProcess(pi.hProcess, 0)
end;

function TDebugger.DelHardwareBreak(const iNumber:integer):Integer;
var
  pr : Boolean;
begin
  result            := 0;
  Cont.ContextFlags := CONTEXT_DEBUG_REGISTERS;
  pr                := GetThreadContext(pi.hThread,cont);
  If (pr) and (iNumber >= 0) and (iNumber <= 3) then
  begin
    case (iNumber) of
      0:  begin
            cont.Dr0:=0;
            Cont.Dr7:= Cont.Dr7 xor BPM0_ENABLED;
          end;
      1:  begin
            cont.Dr1:=0;
            Cont.Dr7:= Cont.Dr7 xor BPM1_ENABLED;
          end;
      2:  begin
            cont.Dr2:=0;
            Cont.Dr7:= Cont.Dr7 xor BPM2_ENABLED;
          end;
      3:  begin
            cont.Dr3:=0;
            Cont.Dr7:= Cont.Dr7 xor BPM3_ENABLED;
          end;
    end;
    BPH[iNumber].addr := 0;
    cont.Dr6          := 0;
    pr                := SetThreadContext(pi.hThread,cont);
    If pr then result:=1;
  end;
end;

function TDebugger.SetHardwareBreak(const iNumber:integer;dwAddress:DWORD):Integer;
var
  pr:Boolean;
  i:integer;
begin
  result:=0;
  For i:=0 to 3 do
    If (BPH[i].addr = dwAddress) then EXIT;
  Cont.ContextFlags:=CONTEXT_DEBUG_REGISTERS;
  pr:=GetThreadContext(pi.hThread,cont);
  If (pr) and (iNumber >= 0) and (iNumber <= 3) and (dwAddress>0) then
  begin
    CASE (iNumber) of
      0:  begin
            cont.Dr0:=dwAddress;
            Cont.Dr7:= Cont.Dr7 or BPM_LOCAL_EXACT or BPM0_ENABLED
          end;
      1:  begin
            cont.Dr1:=dwAddress;
            Cont.Dr7:= Cont.Dr7 or BPM_LOCAL_EXACT or BPM1_ENABLED
          end;
      2:  begin
            cont.Dr2:=dwAddress;
            Cont.Dr7:= Cont.Dr7 or BPM_LOCAL_EXACT or BPM2_ENABLED
          end;
      3:  begin
            cont.Dr3:=dwAddress;
            Cont.Dr7:= Cont.Dr7 or BPM_LOCAL_EXACT or BPM3_ENABLED
          end;
    end;
    BPH[iNumber].addr:=dwAddress;
    cont.Dr6:=0;
    pr:=setThreadContext(pi.hThread,cont);
    If (pr) then result:=1;
  end;
end;

function TDebugger.ReadMemorySelf(dwAddress:DWORD;pBuffer:Pointer;dwSize:dword):integer;
var
  dwRead:DWORD;
  dwOld:DWORD;
begin
  Result:=0;
  If ((VirtualProtectEx(pi.hProcess,Pointer(dwAddress),dwSize,PAGE_READONLY,dwOld)) = FALSE) then exit;
  ReadProcessMemory(pi.hProcess,POINTER(dwAddress),pBuffer,dwSize,dwRead);
  If (VirtualProtectEx(pi.hProcess,Pointer(dwAddress),dwSize,dwOld,dwOld) = FALSE) then exit;
  If (dwRead = dwSize) then Result:=1;
end;

function TDebugger.ReadMemory(const dwAddress: DWORD; const pBuffer: Pointer; const dwSize: dword):integer;
var
  dwRead:DWORD;
begin
  Result := 0;
  ReadProcessMemory(pi.hProcess, POINTER(dwAddress), pBuffer, dwSize, dwRead);
  if (dwRead = dwSize) then Result := 1;
end;

function TDebugger.WriteMemory(dwAddress:DWORD;pBuffer:Pointer;dwSize:dword):integer;
var
  dwWrite:DWORD;
begin
  Result:=0;
  WriteProcessMemory(pi.hProcess,POINTER(dwAddress),pBuffer,dwSize,dwWrite);
  if (dwWrite = dwSize) then Result:=1;
end;

function TDebugger.GetModuleSize(dwModuleHandle: DWord): DWord;
var
  IDH: PImageDosHeader;
  INH: PimageNtHeaders;
begin
  Result := 0;
  if (dwModuleHandle = 0) then Exit;
  IDH := Pointer(dwModuleHandle);
  if isBadReadPtr(IDH,SizeOf(TImageDosHeader)) then Exit;
  if (IDH^.e_magic <> IMAGE_DOS_SIGNATURE) then Exit;
  INH := Pointer(dwModuleHandle+DWord(IDH^._lfanew));
  if (INH^.Signature <> IMAGE_NT_SIGNATURE) then Exit;
  if isBadReadPtr(INH,SizeOf(TImageNtHeaders)) then Exit;
  Result := INH^.OptionalHeader.SizeOfImage;
end;

procedure TDebugger.Disasm(const ADDR, RVA: dword; pDisOpt: PDisasmOptions);
var
  dBuff   : array [0..150] of Char;
  Buff    : array [0..30] of Byte;
  Len     : dword;
  Command : TDisCommand;
  Instr   : TInstruction;
  Options : TMnemonicOptios;
  Cmds    : integer;
begin
  Options.RealtiveOffsets   := False;
  Options.AlternativeAddres := RVA;
  Options.AddAddresPart     := True;
  Options.AddHexDump        := True;
  Options.MnemonicAlign     := 14;
  Cmds                      := 0;

  ZeroMemory(@Instr, SizeOf(TInstruction));
  ZeroMemory(@Command, SizeOf(TDisCommand));
  ZeroMemory(@Buff, SizeOf(Buff));
  ZeroMemory(@DBuff, SizeOf(DBuff));
  ZeroMemory(pDisOpt, SizeOf(pDisOpt^));

  If ReadMemory(ADDR, @Buff, 30) = 1 then
  begin
    Len := InstrDecode(@Buff, @Instr, False);
    InstrDasm(@Instr, @Command, False);
    MakeMnemonic(dBuff, @Command, @Options);
    pDisOpt^.Mnem    := dBuff;
    pDisOpt^.Command := Command;
    pDisOpt^.Len     := Len;
    pDisOpt^.Instr   := Instr;
  end;
end;

Function TDebugger.SetBpx(const iNumber: integer; dwAddress: DWORD):Integer;
var
  dbByte: byte;
begin
  result := 0;
  If (BPX[iNumber].addr = dwAddress) then exit;
  If (iNumber >= 0) and (iNumber <= 15) then
  begin
    ReadMemory(dwAddress,@dbByte,1);
    BPX[iNumber].addr:=dwAddress;
    BPX[iNumber].bb:=dbByte;
    dbByte:=$CC;
    WriteMemory(dwAddress,@dbByte,1);
    result:=1;
    FDebugLog:='SetBpx['+IntTOStr(iNumber)+'] = '+IntToHex(dwAddress,8);
  end;
end;

function TDebugger.DelBpx(const iNumber:integer):Integer;
var
  dbBack:byte;
begin
  result:=0;
  If (iNumber >= 0) and (iNumber <= 15) then
  begin
    dbBack:=BPX[iNumber].bb;
    WriteMemory(BPX[iNumber].addr,@dbBack,1);
    BPX[iNumber].addr:=0;
    BPX[iNumber].bb:=0;
    result:=1;
  end;
end;

function TDebugger.MyBPX(const dwAdress:DWORD): integer;
var
  i:integer;
begin
  Result := -1;
  For i := 0 to 15 do
    If (BPX[i].addr = dwAdress) then
    begin
      Result:=i;
      break;
    end;
end;

function TDebugger.MyBPH(const dwAdress:DWORD): integer;
var
  i:integer;
begin
  Result:=-1;
  For i := 0 to 3 do
    If (BPH[i].addr = dwAdress) then
    begin
      Result:=i;
      break;
    end;
end;

function TDebugger.MyBPM(const dwAdress:DWORD): integer;
var
  i:integer;
begin
  Result:=-1;
  For i := 0 to 15 do
    If (BPM[i].addr >= dwAdress) and (BPM[i].addr <= dwAdress + BPM[i].size) then
    begin
      Result:=i;
      break;
    end;
end;

procedure TDebugger.CreateProcessDebugEvent(const DE : TDebugEvent);
begin
  ImageBase         := DWORD(DE.CreateProcessInfo.lpBaseOfImage);
  EntryPoint        := DWORD(DE.CreateProcessInfo.lpStartAddress);
  PID               := DE.dwProcessId;
  lpThreadLocalBase := DE.CreateProcessInfo.lpThreadLocalBase;

  If (iSetEpBreak = 1)
    then SetBpx(EP_BREAK,EntryPoint);
end;

{procedure TDebugger.GetEax();
type
  PClientID = ^TClientID;
    TClientID = packed record
    UniqueProcess: DWORD;
    UniqueThread:  DWORD;
  end;
  PTHREAD_BASIC_INFORMATION = ^THREAD_BASIC_INFORMATION; 
    THREAD_BASIC_INFORMATION = packed record
    ExitStatus:     DWORD;
    TebBaseAddress: Pointer;
    ClientId:       TClientID;
    AffinityMask:   DWORD;
    Priority:       DWORD;
    BasePriority:   DWORD; 
 end;

 TNtQueryInformationThread = function(hThread: THandle;
    ThreadInformationClass: Integer;
    var ThreadInformation;
    ThreadInformationLength: Integer;
    var ReturnLength: Integer): Integer; stdcall;
const
  ProcessBasicInformation   = $00000000;
  ThreadBasicInformation    = ProcessBasicInformation;
var
  NtQueryInformationThread: TNtQueryInformationThread;
  tbi: THREAD_BASIC_INFORMATION;
  hDup: THandle;
  ret: Integer;
  Status,BytesRead:dword;
  tmp,tmp2,adr: dword;
begin
  @NtQueryInformationThread := GetProcAddress(GetModuleHandle('ntdll.dll'), 'NtQueryInformationThread');
  Status := NtQueryInformationThread(pi.hThread,ThreadBasicInformation,tbi,sizeof(tbi),ret);
  if (status >= 0) then
  begin
    tmp2:=0;
    tmp:=DWORD(tbi.TebBaseAddress)+8;
    ReadProcessMemory(pi.hProcess,Pointer(tmp),@tmp2,4,BytesRead);
    tmp:= (tmp2 and $FFFF0000) + $FD30 + $B0;
    ReadProcessMemory(pi.hProcess,Pointer(tmp),@adr,4,BytesRead);
    adr:=$4131AF;
    WriteProcessMemory(pi.hProcess,Pointer(tmp),@adr,4,BytesRead);

    FDebugLog:='> 111'+IntToHex(adr,8);
  end;
end;}

procedure TDebugger.SystemBreak(const ERec : _EXCEPTION_RECORD);
begin
  DbgControl       := DBG_CONTINUE;
  IsFirstException := FALSE;
  If (iHideDdg = 1) then
  begin
    HideDBG();
  end;
  If (SystemBpHandler > 0) then
  begin
    @CustomHandler := POINTER(SystemBpHandler);
    CustomHandler(ERec);
  end;
end;

procedure TDebugger.ExceptionGuardPage(const ERec : _EXCEPTION_RECORD);
var
  address : DWORD;
  n       : integer;
begin
  DbgControl := DBG_EXCEPTION_NOT_HANDLED;
  address    := DWORD(ERec.ExceptionAddress);
  n          := MyBPM(address);
  If (n > 0) then
  begin
    DelBPM(n);
    DbgControl   := DBG_CONTINUE;
    FDebugLog    := 'Guard Page ['+IntToStr(n)+']: '+IntToHex(address,8);
    FEventReason := BPM_HANDLER;
    If (PageGuardHandler > 0) then
    begin
      @CustomHandler := POINTER(PageGuardHandler);
      CustomHandler(ERec);
    end;
  end;
end;

procedure TDebugger.HandleMyBpx();
begin
  Cont.ContextFlags := CONTEXT_CONTROL;
  GetThreadContext(pi.hThread,cont);
  cont.Eip          := cont.Eip - 1;
  SetThreadContext(pi.hThread,cont);
  DbgControl        := DBG_CONTINUE;
end;

procedure TDebugger.ExceptionSingleStep(const ERec : _EXCEPTION_RECORD);
var
  address : DWORD;
  temp    : DWORD;
  n       : Integer;
begin
  DbgControl := DBG_EXCEPTION_NOT_HANDLED;
  address    := DWORD(ERec.ExceptionAddress);
  n          := MyBPH(address);
  If (n >= 0) then
  begin
    DbgControl   := DBG_CONTINUE;
    FDebugLog    := 'HardwareBreakpoint ['+IntToStr(n)+']: ' + IntToHex(address,8);
    FEventReason := BPH_HANDLER;
    DelHardwareBreak(n);
    exit;
  end;
  If (iOepTrace = 0) and (iTracing = 1) then
  begin
    iTracing     := 0;
    FEventReason := TRACING_HANDLER;
    DbgControl   := DBG_CONTINUE;
    If (SingleStepHandler > 0) then
    begin
      @CustomHandler := POINTER(SingleStepHandler);
      CustomHandler(ERec);
    end;
  end;
  If (iOepTrace = 1) and (iTracing = 1) then
  begin
    DbgControl := DBG_CONTINUE;
    FDebugLog  := 'Tracing ['+IntToStr(n)+']: ' + IntToHex(address,8);
    exit;
  end;
end;

procedure TDebugger.ExceptionBreakPoint(const ERec : _EXCEPTION_RECORD);
var
  address : DWORD;
  n       : Integer;
begin
  DbgControl := DBG_EXCEPTION_NOT_HANDLED;
  address    := DWORD(ERec.ExceptionAddress);
// --- pDbgBreakPoint ---
  If (IsFirstException) then
  begin
    //GetModuleCodeSize()
    //If (address = DWORD(pDbgBreakPoint)) then
    begin
      FEventReason := BPX_SYS_HANDLER;
      SystemBreak(ERec);
      Exit;
    end;
  end;
// ----------------------
  n := MyBPX(address);
  FBpxIndex    := n;
  If (n = EP_BREAK) then
  begin
    FDebugLog    := 'Breakpoint [EP]: ' + IntToHex(address,8);
    Log(FDebugLog);
    FEventReason := EP_HANDLER;
    DelBpx(EP_BREAK);
    HandleMyBpx();
    If (Int3Handler > 0) then
    begin
      @CustomHandler := POINTER(Int3Handler);
      CustomHandler(ERec);
    end;
  end;
// ----------------------
  If (n > 0) then
  begin
    FDebugLog    := 'Breakpoint [' + IntToStr(n) + ']: ' + IntToHex(address,8);
    Log(FDebugLog);
    FEventReason := BPX_HANDLER;
    DelBpx(n);
    HandleMyBpx();
    If (Int3Handler > 0) then
    begin
      @CustomHandler := POINTER(Int3Handler);
      CustomHandler(ERec);
    end;
  end;
end;

function GetDLLName(const hProc: THandle; const DE: TDebugEvent): string;
var
  buf      : array[0..1024] of char;
  pstr     : pointer;
  NumRead  : dword;
  TryRead  : integer;
begin
  Result := '';
  if (DE.LoadDLL.lpImageName = nil) then
  begin
    exit;
  end;
  
  ZeroMemory(@buf, sizeof(buf));
  If (GetModuleFileNameA(DWORD(DE.LoadDll.lpBaseOfDll), @buf, 1024) <> 0) then
  begin
    Result := buf;
    exit;
  end;

  Win32Check(ReadProcessMemory(hProc, DE.LoadDLL.lpImageName, @pstr, sizeof(pstr), NumRead));
  // Assert(NumRead = 4);
  If (NumRead = 4) then
  begin
    if (pstr <> nil) then
    begin
      TryRead := sizeof(buf);
      while (ReadProcessMemory(hProc, pstr, @buf, TryRead, NumRead) = false) do
      begin
         TryRead := TryRead - 1;
      end;
    end;
    if (DE.LoadDLL.fUnicode <> 0) then
      Result := PWideChar(@buf)
    else
      Result := buf;
  end;
end;

function GetDebugString(hProc: THandle; DE: TDebugEvent): string;
var
  buf     : array[0..MAX_PATH] of char;
  NumRead : dword;
begin
  result := '';
  If (DE.DebugString.nDebugStringLength > 0) then
  begin
    Win32Check(ReadProcessMemory(hProc, DE.DebugString.lpDebugStringData, @buf, DE.DebugString.nDebugStringLength, NumRead));
    if (DE.DebugString.fUnicode <> 0) then
      Result := PWideChar(@buf)
    else
      Result := buf;
  end;
end;

procedure TDebugger.ProcessLoadDllDebugEvent(const DE : TDebugEvent);
var
  ibase: dword;
  sDll : string;
begin
  if (iHandleDll = 0) then
  begin
    exit;
  end;

  sDll := GetDLLName(pi.hProcess, DE);
  FDebugLog := ' - DLL_BASE: ' + IntToHex(DWORD(DE.LoadDLL.lpBaseOfDll), 8) + ' DLL_NAME: ' + sDll;
  Log(FDebugLog);

  If (DllHandler > 0) then
  begin

  end;
end;

function GetEventCode(const dwDebugEventCode: dword): string;
begin
  result := 'UNKNOWN';
  case dwDebugEventCode of
    EXCEPTION_DEBUG_EVENT      : result := 'EXCEPTION_DEBUG_EVENT';
    CREATE_THREAD_DEBUG_EVENT  : result := 'CREATE_THREAD_DEBUG_EVENT';
    CREATE_PROCESS_DEBUG_EVENT : result := 'CREATE_PROCESS_DEBUG_EVENT';
    EXIT_THREAD_DEBUG_EVENT    : result := 'EXIT_THREAD_DEBUG_EVENT';
    EXIT_PROCESS_DEBUG_EVENT   : result := 'EXIT_PROCESS_DEBUG_EVENT';
    LOAD_DLL_DEBUG_EVENT       : result := 'LOAD_DLL_DEBUG_EVENT';
    UNLOAD_DLL_DEBUG_EVENT     : result := 'UNLOAD_DLL_DEBUG_EVENT';
    OUTPUT_DEBUG_STRING_EVENT  : result := 'OUTPUT_DEBUG_STRING_EVENT';
    RIP_EVENT                  : result := 'RIP_EVENT';
  end;
end;

procedure TDebugger.ProcessOutputDebugString(const DE : TDebugEvent);
var
  sDebugString : string;
begin
  sDebugString := GetDebugString(pi.hProcess, DE);
  FDebugLog := ' - DEBUG_STRING: ' + sDebugString;
  Log(FDebugLog);
end;

function TDebugger.DoDebugLoop(var dwState, dwReason: dword): dword;
var
  IsExcept : boolean;
begin
  result := 0;
  If (iState = PROCESS_SUSPENDED) then
  begin
    iState := PROCESS_RUNNING;
    ResumeThread(pi.hThread);
  end;
  IsExcept     := FALSE;
  DbgControl   := DBG_CONTINUE;
  FEventReason := 0;
  if WaitForDebugEvent(DE, 0) then
  begin
    result    := 1;
    FDebugLog := 'DebugEvent: ' + GetEventCode(DE.dwDebugEventCode);
    Log(FDebugLog);
    if (DE.dwDebugEventCode = CREATE_PROCESS_DEBUG_EVENT) then
    begin
      CreateProcessDebugEvent(DE);
    end;
    if (DE.dwDebugEventCode = EXIT_THREAD_DEBUG_EVENT) then
    begin

    end;
    if (DE.dwDebugEventCode = CREATE_THREAD_DEBUG_EVENT) then
    begin

    end;
    if (DE.dwDebugEventCode = LOAD_DLL_DEBUG_EVENT) then
    begin
      ProcessLoadDllDebugEvent(DE);
    end;
    if (DE.dwDebugEventCode = UNLOAD_DLL_DEBUG_EVENT) then
    begin

    end;
    if (DE.dwDebugEventCode = OUTPUT_DEBUG_STRING_EVENT) then
    begin
      ProcessOutputDebugString(DE);
    end;
    if (DE.dwDebugEventCode = EXCEPTION_DEBUG_EVENT) then
    begin
      IsExcept  := TRUE;
      FDebugLog := 'Exception Code: '+IntToHex(DE.Exception.ExceptionRecord.ExceptionCode,8) + ' Address: ' + IntToHex(DWORD(DE.Exception.ExceptionRecord.ExceptionAddress),8);
      Log(FDebugLog);
      case DE.Exception.ExceptionRecord.ExceptionCode of
        EXCEPTION_SINGLE_STEP: ExceptionSingleStep(DE.Exception.ExceptionRecord);
        EXCEPTION_BREAKPOINT : ExceptionBreakPoint(DE.Exception.ExceptionRecord);
        EXCEPTION_GUARD_PAGE : ExceptionGuardPage(DE.Exception.ExceptionRecord);
        else DbgControl := DBG_EXCEPTION_NOT_HANDLED;
      end;

    end;
    if (DE.dwDebugEventCode = EXIT_PROCESS_DEBUG_EVENT) then
    begin
      FDebugLog := 'Process Exited';
      Log(FDebugLog);
      iState    := PROCESS_TERMINATED;

      ContinueDebugEvent(DE.dwProcessId, DE.dwThreadId, DBG_TERMINATE_PROCESS);
      CloseHandle(pi.hThread);
      CloseHandle(pi.hProcess);
      exit;
    end;
    If (not IsExcept) then
    begin
      DbgControl := DBG_CONTINUE;
    end;
    ContinueDebugEvent(DE.dwProcessId, DE.dwThreadId, DbgControl);
  end;

  dwState  := iState;
  dwReason := FEventReason;
end;

{function TDebugger.TraceToOep():DWORD;
var
  cmd:integer;
  eip:dword;
  tmp:dword;
begin
  dwOEP := $FFFFFFFF;
  cmd   := 0;
  eip   := MAX_COMMAND_COUNT;
  iOepTrace := 1;
  repeat
    SetTraceMode(1);
    DoDebugLoop();
    inc(cmd);
    tmp := GetEIP();

    If (tmp > $20000000)
      then continue;

    If (tmp < (eip - MAX_COMMAND_COUNT)) then
    begin
      dwOEP := tmp;
      SetTraceMode(0);
    end;
    eip := tmp;
  until (cmd >= MAX_COMMAND_COUNT) or (iState = PROCESS_TERMINATED) or (dwOEP <> $FFFFFFFF);

  iOepTrace := 0;
  If (OepHandler > 0) then
  begin
    @CustomHandler := POINTER(OepHandler);
    CustomHandler(DE.Exception.ExceptionRecord);
  end;
end;}

function TDebugger.RunDebug(const FileName:string): dword;
begin
  iLastError := ERROR_SUCCES;
  Result     := 0;
  iNeedStop  := 0;
  FilePath   := FileName;
  If (Int3Handler > 0) and IsBadCodePtr(POINTER(Int3Handler)) then
  begin
    iLastError := BAD_HANDLER_INT3;
    exit;
  end;
  If (PageGuardHandler > 0) and IsBadCodePtr(POINTER(PageGuardHandler)) then
  begin
    iLastError := BAD_HANDLER_GUARD;
    exit;
  end;
  If (SingleStepHandler > 0) and IsBadCodePtr(POINTER(SingleStepHandler)) then
  begin
    iLastError := BAD_HANDLER_STEP;
    exit;
  end;
  If (SystemBpHandler > 0) and IsBadCodePtr(POINTER(SystemBpHandler)) then
  begin
    iLastError := BAD_HANDLER_SYS;
    exit;
  end;
  If (OepHandler > 0) and IsBadCodePtr(POINTER(OepHandler)) then
  begin
    iLastError := BAD_HANDLER_OEP;
    exit;
  end;
  If (PrepareDebug() = 0) then
  begin
    iLastError := CAN_NOT_CREATE_PROCESS;
    exit;
  end;
  Result := 1;
end;

function TDebugger.DumpProcess():integer;
var
  Buffer:dword;
  FileHandle:dword;
  dwAddress,dwSize,dwTmp:dword;
  DumpName:string;
  i:integer;
begin
  result        := 0;
  dwSize        := $1000;
  Pe.ReadOnly   := TRUE;
  Pe.PEFilePath := FilePath;
  If (PE.PEFilePath = '1') then
  begin
    iLastError := CAN_NOT_OPEN_FILE;
    exit;
  end;
  If (PE.PEFilePath = '2') then
  begin
    iLastError := NOT_VALID_PE_FILE;
    exit;
  end;

  dwSize := PE.GetRealImageSize();
  Pe.ClosePE;

  DumpName := ExtractFilePath(FilePath)+'dump.exe';
  Buffer   := LocalAlloc(lmem_ZeroInit,dwSize);
  If (Buffer = 0) then
  begin
    iLastError := CAN_NOT_GET_MEM;
    exit;
  end;
  If ReadMemory(dwAddress,POINTER(Buffer),dwSize) = 0 then
  begin
    iLastError := CAN_NOT_READ_MEM;
    exit;
  end;
  FileHandle := CreateFile(pchar(DumpName),$C0000000,$3,nil,$2,$80,$0);
  If (FileHandle = DWORD(-1)) then
  begin
    iLastError := CAN_NOT_CREATE_FILE;
    LocalFree(Buffer);
    exit;
  end;
  If (not WriteFile(FileHandle,POINTER(Buffer),dwSize,dwTmp,nil)) then
  begin
    iLastError := CAN_NOT_WRITE_TO_FILE;
    LocalFree(Buffer);
    exit;
  end;
  LocalFree(Buffer);  
  result:=1;
end;

procedure TDebugger.TerminateDebug();
begin
  iNeedStop := 1;
  FilePath  := '';
  ProcessTerminate();
end;

end.
